
[Setting category="General" name="Show in Editor" description="Show in any mode where you are not driving, such as map editor or replay editor"]
bool showEditor = true;

uint blocks_size = 0;
string lastBlockId = '';
uint items_size = 0;
string lastItemId = '';

bool IsCheckpoint(CGameCtnBlock@ block) {
    return block.WaypointSpecialProperty !is null;
}
bool IsCheckpoint(CGameCtnAnchoredObject@ item) {
    return item.WaypointSpecialProperty !is null;
}

void Update_InEditor() {
    CGameCtnChallenge@ rootMap = cast<CGameCtnChallenge>(GetApp().RootMap);

    if ((!inEditor || hasChanged())) {
        // keep the previously-determined CP data, unless in the map editor

        reset(); // won't be called if returning from map to editor

        inEditor = true;
        curMapId = rootMap.IdName;
        maptype = rootMap.MapType;
        string comment = rootMap.Comments;

        blocks_size = rootMap.Blocks.Length;
        if (blocks_size > 0) {
            lastBlockId = rootMap.Blocks[rootMap.Blocks.Length-1].IdName;
        } else {
            lastBlockId = '';
        }
        items_size = rootMap.AnchoredObjects.Length;
        if (items_size > 0) {
            lastItemId = rootMap.AnchoredObjects[rootMap.AnchoredObjects.Length-1].IdName;
        } else {
            lastItemId = '';
        }

        CpData@[] new_cps = {};
        // Start should always have priority over multilap (even if multilap has number and start doesn't)
        int start = -1;
        bool startIsItem = false;
        for(uint i = 0; i < rootMap.Blocks.Length; i++) {
            if (rootMap.Blocks[i].WaypointSpecialProperty !is null) {

                if ((rootMap.Blocks[i].WaypointSpecialProperty.Tag == 'Spawn')) {
                    start = i;
                } else {
                    if (rootMap.Blocks[i].WaypointSpecialProperty.Tag == "StartFinish") {
                        StartOrder = rootMap.Blocks[i].WaypointSpecialProperty.Order; // will be overwritten if any start block
                    }
                    if (rootMap.Blocks[i].WaypointSpecialProperty.Tag == 'Checkpoint' || rootMap.Blocks[i].WaypointSpecialProperty.Tag == 'LinkedCheckpoint') {
                        if (rootMap.Blocks[i].WaypointSpecialProperty.Order == 0) {
                            unordered++;
                        } else {
                            CPorders.InsertLast(rootMap.Blocks[i].WaypointSpecialProperty.Order);
                        }
                    }
                }
                new_cps.InsertLast(CpData(rootMap.Blocks[i]));
                if (Time::get_Now() - startTime > 20) {
                    yield();
                    startTime = Time::get_Now();
                    if (rootMap is null || GetApp().RootMap !is rootMap) {
                        reset();
                        return;}
                }
            }
        }
        for(uint i = 0; i < rootMap.AnchoredObjects.Length; i++) {
            if (rootMap.AnchoredObjects[i].WaypointSpecialProperty !is null) {

                if ((rootMap.AnchoredObjects[i].WaypointSpecialProperty.Tag == 'Spawn')) {
                    start = i;
                    startIsItem = true;
                } else {
                    if (rootMap.AnchoredObjects[i].WaypointSpecialProperty.Tag == "StartFinish") {
                        StartOrder = rootMap.AnchoredObjects[i].WaypointSpecialProperty.Order; // will be overwritten if any start block

                    }
                    if (rootMap.AnchoredObjects[i].WaypointSpecialProperty.Tag == 'Checkpoint' || rootMap.AnchoredObjects[i].WaypointSpecialProperty.Tag == 'LinkedCheckpoint') {
                        if (rootMap.AnchoredObjects[i].WaypointSpecialProperty.Order == 0) {
                            unordered++;
                        } else {
                            CPorders.InsertLast(rootMap.AnchoredObjects[i].WaypointSpecialProperty.Order);
                        }
                    }
                }
                new_cps.InsertLast(CpData(rootMap.AnchoredObjects[i]));
                if (Time::get_Now() - startTime > 20) {
                    yield();
                    startTime = Time::get_Now();
                    if (rootMap is null || GetApp().RootMap !is rootMap) {
                        reset();
                        return;}
                }
            }
        }

        CPorders.SortAsc();
        if (startIsItem) {
            if (start != -1 && uint(start) < rootMap.AnchoredObjects.Length && rootMap.AnchoredObjects[start].WaypointSpecialProperty !is null && rootMap.AnchoredObjects[start].WaypointSpecialProperty.Tag == 'Spawn'){
                StartOrder = rootMap.AnchoredObjects[start].WaypointSpecialProperty.Order;
            }
        } else {
            if (start != -1 && uint(start) < rootMap.Blocks.Length && rootMap.Blocks[start].WaypointSpecialProperty !is null && rootMap.Blocks[start].WaypointSpecialProperty.Tag == 'Spawn'){
                StartOrder = rootMap.Blocks[start].WaypointSpecialProperty.Order;
            }
        }
        if (comment.ToLower().Contains('/uci order') || comment.ToLower().Contains('/uci lowest')) {
            if (StartOrder < 65535) {StartOrder = 65535;}
            StartOrder = ((StartOrder - 65535) | Options::Lowest) + 65535;
        } else if (comment.ToLower().Contains('/uci hide')) {
            if (StartOrder < 65535) {StartOrder = 65535;}
            StartOrder = ((StartOrder - 65535) | Options::Hide) + 65535;
        } else if (comment.ToLower().Contains('/uci xdd')) {
            if (StartOrder < 65535) {StartOrder = 65535;}
            StartOrder = ((StartOrder - 65535) | Options::xdd) + 65535;
        } else if (comment.ToLower().Contains('/uci always')) {
            if (StartOrder < 65535) {StartOrder = 65535;}
            StartOrder = ((StartOrder - 65535) | Options::Always) + 65535;
        }
        cps = new_cps;
        partialReset();
        setCurrentMapSettings();
    }
}


bool hasChanged() {
    CGameCtnChallenge@ rootMap = cast<CGameCtnChallenge>(GetApp().RootMap);
    if (rootMap is null) {return true;}
    if (GetApp().Editor is null || cast<CGameCtnEditorFree>(GetApp().Editor) is null) {return false;}
    return !(blocks_size == rootMap.Blocks.Length && lastBlockId == rootMap.Blocks[rootMap.Blocks.Length-1].IdName && items_size == rootMap.AnchoredObjects.Length && ((rootMap.AnchoredObjects.Length == 0 && lastItemId == '') || (rootMap.AnchoredObjects.Length > 0 && lastItemId == rootMap.AnchoredObjects[rootMap.AnchoredObjects.Length-1].IdName)));
}
