/*
 * by ArEyeses
 * (some code copied with permission from CP counter by Phlarx)
 */

/* there are some parts that are commented out because I can't find a way to make them work
 * - Type of checkpoint block e.g. ring (I can find this, but can't tell which block/item is which checkpoint)
 * - rotation of block when in freemode (also needs above point to be useful)
 */

[Setting category="General" name="Show checkpoint indicator"]
bool show = true;

[Setting category="General" name="Track CPs when hidden" description="Count checkpoints when not shown, allowing you to enable when you're trying to find the last cp"]
bool run = true;

[Setting category="General" name="Dot size" min=2 max=64]
int dotSize = 7;

[Setting category="General" name="Show when UI hidden"]
bool showUIHidden = true;

[Setting category="General" name="Show when pause menu open (*pause menu detection unreliable)"]
bool showPause = true;

[Setting category="General" name="Smaller dots with distance" description="Further checkpoints have smaller dots. Will never become too small to see"]
bool showDistance = true;

[Setting category="General" name="Single-color dots (small rendering speed increase)"]
bool lowDrawQuality = false;

// using Options enum; this means you can disable the map options later
uint currentMapSettings = 0;
uint numMapSettings = 0;

/**
 * If true, then this plugin has detected that we are on a map.
 * If this and inEditor false, none of the other values are valid.
 */
bool inMap = false;
// editor version of inMap
bool inEditor = false;
// if the car is actually driving (not in an intro/finished)
bool inGame = false;

/**
 * The ID of the map whose checkpoints are currently in the arraya.
 */
string curMapId = "";

/**
 * The current respawn location (thus the most recent cp taken).
 * when this changes, it means we've taken a new cp
 */
uint preCPIdx = uint(-2);

// the order of the start block. If mapper sets this to __, will only show checkpoints in order
uint StartOrder = 0;

// number of checkpoints with no order
uint unordered = 0;
// orders of checkpoints used in the map (for if they will be displayed in order). Does contain duplicates
array<uint> CPorders = {};
// orders of checkpoints that haven't been taken (only used to filter display)
array<uint> NextCPorders = {};

string maptype = "";

uint64 startTime;

enum Options {
    // only show the checkpoint with lowest remaining order
    Lowest = 0b1,
    // hide all checkpoints (except those with 'Always')
    Hide = 0b10,
    xdd = 0b100,
    Always = 0b1000,
}
bool isOption(uint order, uint option) {
    return (order > 65535 && (order - 65535)&option != 0);
}

void OnSettingsChanged() {
    for (uint i = 0; i < cps.Length; i++) {
        cps[i].setColor();
    }
    setCurrentMapSettings();
}


void reset() {//resets the current map and checkpoint lists
    curMapId = "";
    nottaken = {};
    cps = {};
    inGame = false;
    inMap = false;
    inEditor = false;
    StartOrder = 0;
    unordered = 0;
    CPorders = {};
    NextCPorders = {};
    maptype = "";
    currentMapSettings = 0;
}
void partialReset() {
    inGame = false;
    if (maptype == "TrackMania\\TM_Royal") {
        StartOrder = 0; // multiple starts, so will be set once you've spawned from one of them
        NextCPorders = CPorders;
    } else if (maptype == "FlagRushArena") {
        StartOrder = 0; // multiple starts, so will be set once you've spawned from one of them
    } else {
        NextCPorders = CPorders;
    }
    nottaken = cps;
}

void Main() {
    while (true) {
        yield();
        startTime = Time::get_Now();

        if (!run && !show) {continue;}
        
        if (GetApp().RootMap is null) {
            reset();
            continue;
        }


        CSmArenaClient@ playground = cast<CSmArenaClient>(GetApp().CurrentPlayground);
        if(playground !is null
            && playground.Arena !is null) {
            Update_InMap();
            continue;
        }

        // playground is null when loading in to map, so need to check editor exists (could be map or replay)
        if (GetApp().Editor is null) {
            reset();
            continue;
        } else if (showEditor) {
            Update_InEditor();
        }
        // continue;
    }
}

void setCurrentMapSettings() {
    currentMapSettings = 0;
    CGameCtnChallenge@ rootMap = cast<CGameCtnChallenge>(GetApp().RootMap);
    if (rootMap is null) {return;}
    string comment = rootMap.Comments;

    if (false) {
        currentMapSettings |= Options::Lowest;
    } else if (isOption(StartOrder, Options::Lowest)) {
        currentMapSettings |= Options::Lowest;
        numMapSettings += 1;
    }
    if (!show) {
        currentMapSettings |= Options::Hide;
    } else if (isOption(StartOrder, Options::Hide)) {
        currentMapSettings |= Options::Hide;
        numMapSettings += 1;
    }
    if (xddMode) {
        currentMapSettings |= Options::xdd;
    } else if (isOption(StartOrder, Options::xdd)) {
        currentMapSettings |= Options::xdd;
        numMapSettings += 1;
    }
    if (false) {
        currentMapSettings |= Options::Always;
    } else if (isOption(StartOrder, Options::Always)) {
        currentMapSettings |= Options::Always;
        numMapSettings += 1;
    }
}

void Update_InMap() {
    CGameCtnChallenge@ rootMap = cast<CGameCtnChallenge>(GetApp().RootMap);
    CSmArenaClient@ playground = cast<CSmArenaClient>(GetApp().CurrentPlayground);

    MwFastBuffer<CGameScriptMapLandmark@> landmarks = playground.Arena.MapLandmarks;

    if(!inMap && (curMapId != rootMap.IdName || GetApp().Editor !is null)) {
        // keep the previously-determined CP data, unless in the map editor

        reset(); // won't be called before if in editor

        inMap = true;
        curMapId = rootMap.IdName;
        maptype = rootMap.MapType;
        string comment = rootMap.Comments;

        int start = -1;
        for(uint i = 0; i < landmarks.Length; i++) {

            if ((landmarks[i].Waypoint is null)) {
                start = i;
            } else {
                if (landmarks[i].Waypoint.IsMultiLap) {
                    StartOrder = landmarks[i].Order; // will be overwritten if any start block
                    
                }
                if (landmarks[i].Tag == 'Checkpoint' || landmarks[i].Tag == 'LinkedCheckpoint') {
                    if (landmarks[i].Order == 0) {
                        unordered++;
                    } else {
                        CPorders.InsertLast(landmarks[i].Order);
                    }
                }
            }
            cps.InsertLast(CpData(landmarks[i]));
        }
        CPorders.SortAsc();
        if (start != -1){
            StartOrder = landmarks[start].Order;
        }
        if (comment.ToLower().Contains('/uci order') || comment.ToLower().Contains('/uci lowest')) {
            if (StartOrder < 65535) {StartOrder = 65535;}
            StartOrder = ((StartOrder - 65535) | Options::Lowest) + 65535;
        }
        if (comment.ToLower().Contains('/uci hide')) {
            if (StartOrder < 65535) {StartOrder = 65535;}
            StartOrder = ((StartOrder - 65535) | Options::Hide) + 65535;
        }
        if (comment.ToLower().Contains('/uci xdd')) {
            if (StartOrder < 65535) {StartOrder = 65535;}
            StartOrder = ((StartOrder - 65535) | Options::xdd) + 65535;
        }
        if (comment.ToLower().Contains('/uci always')) {
            if (StartOrder < 65535) {StartOrder = 65535;}
            StartOrder = ((StartOrder - 65535) | Options::Always) + 65535;
        }

        partialReset();
        setCurrentMapSettings();
    }

    if (playground.Map is null
        || playground.GameTerminals.Length <= 0
        || playground.GameTerminals[0].UISequence_Current != CGamePlaygroundUIConfig::EUISequence::Playing
        || cast<CSmPlayer>(playground.GameTerminals[0].GUIPlayer) is null) {
        partialReset();
        return;
    }

    auto player = cast<CSmPlayer>(playground.GameTerminals[0].GUIPlayer);

    // not sure what this means, since test in editor is uint(-2)
    if(player.CurrentLaunchedRespawnLandmarkIndex == uint(-1)) {
        partialReset();
        return;
    }

    if (!inGame) {
        inGame = true;
        preCPIdx = uint(-2); // this will never match current CP, so will always re-check CPs
    }

    if(preCPIdx != player.CurrentLaunchedRespawnLandmarkIndex && landmarks.Length > player.CurrentLaunchedRespawnLandmarkIndex) {
        // taken new CP
        preCPIdx = player.CurrentLaunchedRespawnLandmarkIndex;

        if (maptype == "FlagRushArena") {
            if (landmarks[preCPIdx].Tag == "SpawnTeam1") {
                StartOrder = 1;
            } else if (landmarks[preCPIdx].Tag == "SpawnTeam2") {
                StartOrder = 2;
            } // else if (landmarks[preCPIdx].Tag == 'Unused Checkpoint') {
            //     StartOrder = 0; // both finishes will be shown
            // } else {
            //     print("unknown start tag [flag rush] " + landmarks[preCPIdx].Tag);
            // }
        } else {
            if(landmarks[preCPIdx].Waypoint is null || landmarks[preCPIdx].Waypoint.IsFinish || landmarks[preCPIdx].Waypoint.IsMultiLap) {
                // if null, it's a start block. if the other flags, it's either a multilap or a finish. In all such cases, reset the nottaken cps to all cps.
                if (maptype == "TrackMania\\TM_Royal") {
                    if (landmarks.Length > player.CurrentLaunchedRespawnLandmarkIndex) { // not editor-spawn without CP
                        StartOrder = landmarks[preCPIdx].Order; // it might be a finish but that doesn't matter, order should still be the same
                    }
                }

                NextCPorders = CPorders;
                nottaken = {};
                nottaken.Reserve(cps.Length - 1); // if multilap, this will be extended to 1 more
                for (uint i = 0; i < cps.Length; i++) {
                    if (cps[i].waypointType != WaypointType::Start) {
                        nottaken.InsertLast(cps[i]);
                    }
                }

            } else {
                // Checkpoint
                array<uint> toremove = {};
                uint value = landmarks[preCPIdx].MarkerId.Value;
                for (uint i = 0; i < nottaken.Length; i++){
                    if (nottaken[i].Value == value) {
                        toremove.InsertLast(i);
                    } else if (landmarks[preCPIdx].Tag == 'LinkedCheckpoint' && nottaken[i].Tag == 'LinkedCheckpoint' && nottaken[i].Order == landmarks[preCPIdx].Order) {
                        toremove.InsertLast(i);
                    }
                }
                toremove.SortDesc(); // toremove is a list of checkpoint indexes to remove from nottaken, in order from last to first
                for (uint i = 0; i < toremove.Length; i++) {
                    if (nottaken[toremove[i]].Order != 0) {
                        uint j = 0;
                        for (; j < NextCPorders.Length; j++) {
                            if (NextCPorders[j] == nottaken[toremove[i]].Order) {break;}
                        }
                        if (j <= NextCPorders.Length) {
                            NextCPorders.RemoveAt(j);}
                        else {
                            print('cp orders error (I think I\'ve fixed this)');
                        }
                    }
                    if (toremove[i] >= nottaken.Length) {
                        print('trying to remove CP after end of not taken');
                        print(toremove[i] + " " + toremove.Length + " " + nottaken.Length);
                    } else {
                        nottaken.RemoveAt(toremove[i]);
                    }
                }
            }
        }
    }
}

